/*
 * Cresques Mapping Suite. Graphic Library for constructing mapping applications.
 *
 * Copyright (C) 2004-5.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 * cresques@gmail.com
 */
package org.cresques.io;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import org.cresques.geo.ViewPortData;
import org.cresques.px.PxRaster;


/**
 * Clase abstracta de la que heredan los drivers de escritura. Tiene los
 * mtodos abstractos que debe implementar cualquier driver de escritura
 * y las funcionalidades y opciones soportadas comunes a todos ellos.
 * @author Nacho Brodin (brodin_ign@gva.es)
 */
public abstract class GeoRasterWriter {
    
	public static int			blockSizeDefault = 256;
	
	private static TreeMap 		supportedExtensions = null;
    protected static TreeMap 	typeList = new TreeMap();
    protected String 			outFileName = null;
    protected String 			inFileName = null;
    protected int 				sizeWindowX = 0;
    protected int 				sizeWindowY = 0;
    protected int 				ulX = 0;
    protected int 				ulY = 0;
    protected PxRaster 			currentRaster;
    protected IDataWriter 		dataWriter = null;
    protected int 				nBands = 0;
    protected String 			ident = null;
    protected String 			driver = null;
    
    /**
     * Registra un formato de escritura
     * @param ext        Extensin del fichero registrado
     * @param clase        Clase que maneja el formato registrado
     */
    public static void registerWriterExtension(String ext, Class clase) {
        if (supportedExtensions == null) {
            supportedExtensions = new TreeMap();
        }

        ext = ext.toLowerCase();
        System.out.println("Write RASTER: extension '" + ext + "' supported.");
        supportedExtensions.put(ext, clase);
    }

    /**
     * Obtiene la lista de extensiones registradas
     * @return Lista de extensiones registradas o null si no hay ninguna
     */
    public static String[] getDriversExtensions(){
    	if (supportedExtensions.size() == 0)
    		return null;
    	String[] list = new String[supportedExtensions.size()];
    	Set values = supportedExtensions.entrySet();
    	int i = 0;
    	for (Iterator it=values.iterator(); it.hasNext(); ) {
            list[i] = (String)((Map.Entry)it.next()).getKey();
            i++;
        }
    	
    	return list;
    }
    
    /**
     * Obtiene la lista de tipos de driver
     * @return Lista de tipos de driver registradas o null si no hay ninguno
     */
    public static String[] getDriversType(){
    	if (typeList.size() == 0)
    		return null;
    	String[] list = new String[typeList.size()];
    	Set values = typeList.entrySet();
    	int i = 0;
    	for (Iterator it=values.iterator(); it.hasNext(); ) {
            list[i] = (String)((Map.Entry)it.next()).getValue();
            i++;
        }
    	
    	return list;
    }
    
    /**
     * Obtiene el tipo de driver a partir de la extensin
     * @param ext	Extensin
     * @return	Tipo
     */
    public static String getDriverType(String ext){
    	return (String)typeList.get(ext);
    }
    
    /**
     * Devuelve el nmero de drivers soportados
     * @return Nmero de drivers soportados
     */
    public static int getNDrivers() {
        return supportedExtensions.size();
    }
    
    /**
     * Devuelve el nmero de tipos de driver registrados
     * @return Nmero de tipos de driver soportados
     */
    public static int getNTypes() {
        return typeList.size();
    }

    /**
     * Devuelve el identificador del driver
     * @return        Identificador del driver
     */
    public String getIdent() {
        return ident;
    }

    /**
     * Obtiene el nombre del driver.
     * @return        Nombre del driver
     */
    public String getDriverName() {
        return driver;
    }

    /**
     * 
     * @return
     */
    public String getDriverType() {
        return driver;
    }
    
    /**
	 * Factoria para obtener escritores de los distintos tipos de raster.
	 * 
	 * @param fName Nombre del fichero.
	 * @return GeoRasterWriter, o null si hay problemas.
	 */
	public static GeoRasterWriter getWriter(String fName) {
		String ext = fName.toLowerCase().substring(fName.lastIndexOf('.')+1);
		GeoRasterWriter grw = null;
		
		if (!supportedExtensions.containsKey(ext)) 
			return grw;
		
		Class clase = (Class) supportedExtensions.get(ext);
		Class [] args = {String.class};
		try {
			Constructor hazNuevo = clase.getConstructor(args);
			Object [] args2 = {fName};
			grw = (GeoRasterWriter) hazNuevo.newInstance(args2);
		} catch (SecurityException e) {
			e.printStackTrace();
		} catch (NoSuchMethodException e) {
			e.printStackTrace();
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
		}
		return grw;
	}
	
	/**
	 * Factoria para obtener escritores de los distintos tipos de raster.
	 * 
	 * @param fName Nombre del fichero.
	 * @return GeoRasterWriter, o null si hay problemas.
	 */
	public static GeoRasterWriter getWriter(IDataWriter dataWriter, 
     										String outFileName, 
     										int blockSize, 
     										int nBands,
     										ViewPortData vp,
     										int compresion,
     										int outSizeX,
     										int outSizeY) {
		String ext = outFileName.toLowerCase().substring(outFileName.lastIndexOf('.')+1);
		GeoRasterWriter grw = null;
		
		if (!supportedExtensions.containsKey(ext)) 
			return grw;
		
		Class clase = (Class) supportedExtensions.get(ext);
		Class [] args = {IDataWriter.class, String.class, Integer.class, Integer.class, 
						 ViewPortData.class, Integer.class, Integer.class, Integer.class};
		try {
			Constructor hazNuevo = clase.getConstructor(args);
			Object [] args2 = {dataWriter, outFileName, new Integer(blockSize), new Integer(nBands), 
								vp, new Integer(compresion), new Integer(outSizeX), new Integer(outSizeY)};
			grw = (GeoRasterWriter) hazNuevo.newInstance(args2);
		} catch (SecurityException e) {
			e.printStackTrace();
		} catch (NoSuchMethodException e) {
			e.printStackTrace();
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
		}
		return grw;
	}
	
    /**
     * Asigna propiedades al driver a partir de un vector de
     * strings donde cada elemento tiene la estructura de
     * propiedad=valor.
     * @param props        Propiedades
     */
    public abstract void setProps(String[] props);

    /**
     * Realiza la funcin de compresin a partir de un GeoRasterFile.
     * @throws IOException
     */
    public abstract void fileWrite() throws IOException;

    /**
     * Realiza la funcin de compresin a partir de los datos pasados por el cliente.
     * @throws IOException
     */
    public abstract void dataWrite() throws IOException;

    /**
     * Cierra el driver
     */
    public abstract void writeClose();
    
    /**
     * Cancela el grabado de datos
     */
    public abstract void writeCancel();

    /**
     * Devuelve la configuracin de la ventana de dialogo
     * para las propiedades del driver de escritura que se est tratando.
     * @return Texto XML con las propiedades
     */
    public abstract String getXMLPropertiesDialog();
}
