/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2004 IVER T.I. and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *  Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *      +34 963862235
 *   gvsig@gva.es
 *      www.gvsig.gva.es
 *
 *    or
 *
 *   IVER T.I. S.A
 *   Salamanca 50
 *   46005 Valencia
 *   Spain
 *
 *   +34 963163400
 *   dac@iver.es
 */
package com.iver.cit.gvsig.fmap.layers;

import java.io.File;
import java.io.IOException;

import com.hardcode.driverManager.DriverLoadException;
import com.hardcode.gdbms.engine.data.DataSource;
import com.hardcode.gdbms.engine.data.DataSourceFactory;
import com.hardcode.gdbms.engine.data.NoSuchTableException;
import com.hardcode.gdbms.engine.data.driver.DriverException;
import com.hardcode.gdbms.engine.data.driver.ObjectDriver;
import com.hardcode.gdbms.engine.values.Value;
import com.iver.cit.gvsig.fmap.core.DefaultFeature;
import com.iver.cit.gvsig.fmap.core.IFeature;
import com.iver.cit.gvsig.fmap.core.IGeometry;
import com.iver.cit.gvsig.fmap.drivers.DriverIOException;
import com.iver.cit.gvsig.fmap.drivers.ExternalData;
import com.iver.cit.gvsig.fmap.drivers.VectorialFileDriver;


/**
 * Adapta un driver de fichero vectorial a la interfaz vectorial, manteniendo
 * adems el estado necesario por una capa vectorial de fichero (el nombre del
 * fichero)
 */
public class VectorialFileAdapter extends VectorialAdapter {
	private boolean driverInitialized = false;
	private File file;
	private SelectableDataSource ds;
	private String dataSourceName;

	/**
	 * <code>reference_count</code> lleva un contador de referencias a este
	 * adaptador, de forma que si la cuenta de referencias no es cero, no
	 * abrimos otra vez el adaptador porque se supone que est abierto.
	 */
	private int reference_count = 0;

	/**
	 * Crea un nuevo VectorialFileAdapter.
	 *
	 * @param file Fichero.
	 */
	public VectorialFileAdapter(File file) {
		this.file = file;
	}

	/**
	 * Devuelve driver.
	 *
	 * @return VectorialFileDriver.
	 */
	VectorialFileDriver getFileDriver() {
		return (VectorialFileDriver) getDriver();
	}

	/**
	 * incrementa el contador de las veces que se ha abierto el fichero.
	 * Solamente cuando el contador est a cero pide al driver que abra el
	 * fichero
	 *
	 * @throws DriverIOException
	 */
	public synchronized void start() throws DriverIOException {
		try {
		    if (reference_count == 0)
		    {
				getFileDriver().open(file);

				if (!driverInitialized) {
					getFileDriver().initialize();
					driverInitialized = true;
				}
		    }
			reference_count++;

		} catch (IOException e) {
			throw new DriverIOException(e);
		}
	}

	/**
	 * decrementa el contador de nmero de aperturas y cuando llega a cero pide
	 * al driver que cierre el fichero
	 *
	 * @throws DriverIOException
	 */
	public synchronized void stop() throws DriverIOException {
		try {
		    if (reference_count == 0)
		    {
		        getFileDriver().close();
		    }
		    else
		        if (reference_count < 0)
		            throw new RuntimeException("Contador de referencias de driver ="
		                    + reference_count + ". Demasiados stop().");
	    	reference_count--;
		} catch (IOException e) {
			throw new DriverIOException(e);
		}
	}

	/**
	 * Is synchronized to allow thread safe access to features stored
	 * in files.
	 * 
	 * @see com.iver.cit.gvsig.fmap.layers.ReadableVectorial#getShape(int)
	 */
	public synchronized IGeometry getShape(int index) throws DriverIOException {
		try {
			return getFileDriver().getShape(index);
		} catch (IOException e) {
			throw new DriverIOException(e);
		}
	}

	/**
	 * @see com.iver.cit.gvsig.fmap.layers.ReadableVectorial#getShapeType()
	 */
	public int getShapeType() throws DriverIOException {
		return getFileDriver().getShapeType();
	}

	/**
	 * @see com.iver.cit.gvsig.fmap.layers.VectorialAdapter#getRecordset()
	 */
	public SelectableDataSource getRecordset() throws DriverLoadException {
		try {
			if (ds == null) {
				VectorialFileDriver driver = (VectorialFileDriver) getDriver();

				if (driver instanceof ExternalData) {
					ExternalData ed = (ExternalData) driver;
					File dataFile = ed.getDataFile(file);
					String driverName = ed.getDataDriverName();

					String name = LayerFactory.getDataSourceFactory().addFileDataSource(driverName,
						dataFile.getAbsolutePath());
					ds = new SelectableDataSource(LayerFactory.getDataSourceFactory().createRandomDataSource(name, DataSourceFactory.AUTOMATIC_OPENING));
				} else if (driver instanceof ObjectDriver) {
					String name = LayerFactory.getDataSourceFactory().addDataSource((ObjectDriver)driver);
					ds = new SelectableDataSource(LayerFactory.getDataSourceFactory().createRandomDataSource(name, DataSourceFactory.AUTOMATIC_OPENING));
				} else {
					return null;
				}
			}
		} catch (NoSuchTableException e) {
			throw new RuntimeException(
				"Error de implementacin, se ha aadido una tabla y luego esa tabla no ha podido ser cargada");
		} catch (DriverException e) {
			throw new RuntimeException(e);
		}

		return ds;
	}

	/**
	 * Devuelve el fichero.
	 *
	 * @return Fichero.
	 */
	public File getFile() {
		return file;
	}
	/**
	 * Returns the feature whose index is numReg
	 * <br>
	 * Is synchronized to do thread safe accessing to features
	 * stored in files.
	 * @param numReg index of feature
	 * @return feature
	 * 
	 */
    public synchronized IFeature getFeature(int numReg) throws com.iver.cit.gvsig.fmap.DriverException
    {
        IGeometry geom;
        IFeature feat = null;
        try {
            geom = getShape(numReg);
            DataSource rs = getRecordset();
            Value[] regAtt = new Value[rs.getFieldCount()];
            for (int fieldId=0; fieldId < rs.getFieldCount(); fieldId++ )
            {
                regAtt[fieldId] =  rs.getFieldValue(numReg, fieldId);
            }

            feat = new DefaultFeature(geom, regAtt, "" + numReg);
        } catch (DriverIOException e) {
            throw new com.iver.cit.gvsig.fmap.DriverException(e);
        } catch (DriverLoadException e) {
            throw new com.iver.cit.gvsig.fmap.DriverException(e);
        } catch (com.hardcode.gdbms.engine.data.driver.DriverException e) {
            throw new com.iver.cit.gvsig.fmap.DriverException(e);
        }
        return feat;
    }

}
