package org.gvsig.remoteClient.gml;

import java.awt.geom.Rectangle2D;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Properties;

import org.gvsig.remoteClient.gml.factories.FeaturesParserFactory;
import org.gvsig.remoteClient.gml.factories.IGeometriesFactory;
import org.gvsig.remoteClient.gml.factories.XMLParserFactory;
import org.gvsig.remoteClient.gml.schemas.XMLSchemaParser;
import org.gvsig.remoteClient.gml.schemas.XMLSchemasManager;
import org.kxml2.io.KXmlParser;
import org.xmlpull.v1.XmlPullParserException;

/* gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2004 IVER T.I. and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *  Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *      +34 963862235
 *   gvsig@gva.es
 *      www.gvsig.gva.es
 *
 *    or
 *
 *   IVER T.I. S.A
 *   Salamanca 50
 *   46005 Valencia
 *   Spain
 *
 *   +34 963163400
 *   dac@iver.es
 */
/* CVS MESSAGES:
 *
 * $Id: GMLReader.java,v 1.1.2.2 2006/11/17 11:38:03 ppiqueras Exp $
 * $Log: GMLReader.java,v $
 * Revision 1.1.2.2  2006/11/17 11:38:03  ppiqueras
 * Corregidos bugs y añadida nueva funcionalidad. Del HEAD.
 *
 * Revision 1.2  2006/11/06 12:15:11  jorpiell
 * Aadido un constructor con un parmetro booleano que se usar para abrir gml's sin validar el esquema
 *
 * Revision 1.1  2006/08/10 12:00:49  jorpiell
 * Primer commit del driver de Gml
 *
 *
 */
/**
 * This class must be used to read any GML 2.x file (2.0, 2.1.1 
 * 2.1.2). It is able to parse GML features that have a xml complex
 * type with only one level of complexity. More levels and GML 3.x 
 * will be supported in next versions.
 * 
 * Name: OpenGIS Geography Markup Language (GML) Implementation Specification,
 * Version: 2.x
 * Project Document: 01-029 (2.0), 02-099 (2.1.1) and 02-069 (v2.1.2)
 * 
 * @author Jorge Piera Llodr (piera_jor@gva.es)
 */
public class GMLReader {
	private XMLSchemasManager schemasManager = null;
	private GMLFeaturesParser featuresParser = null;
	private File m_File = null;
	
	/**
	 * Constructor
	 * @param file
	 * GML File to read
	 * @param factory
	 * Geometries factory that is used to create the parsed 
	 * GML geometries
	 * @throws Exception 
	 * @throws FileNotFoundException
	 * When the file is not found
	 */
	public GMLReader(File file,IGeometriesFactory factory) throws Exception{
		this.m_File = file;
		initialize(factory);
	}
	
	/**
	 * Constructor
	 * @param file
	 * GML File to read
	 * @param factory
	 * Geometries factory that is used to create the parsed 
	 * GML geometries
	 * @param mustValidate
	 * To validate or not validate the schemas
	 * @throws Exception 
	 * @throws FileNotFoundException
	 * When the file is not found
	 */
	public GMLReader(File file,IGeometriesFactory factory,boolean mustValidate) throws Exception{
		this(file,factory);
	}
	
	/**
	 * Initializes the XMLSchemaParser, tries to parse  the GML header 
	 * to obtain the GML version and parses all the needed schemas
	 * @param factory
	 * Geometries factory that is used to create the parsed 
	 * GML geometries
	 * @throws Exception 
	 */
	private void initialize(IGeometriesFactory factory) throws Exception{
		XMLSchemaParser parser = new XMLParserFactory().createSchemaParser(m_File);
		
		//Retrieve the header attributes and download and parse the schemas
		schemasManager = new XMLSchemasManager(m_File);
		schemasManager.parse(parser);
		
		//Parse the GML global fields
		featuresParser = new FeaturesParserFactory().createParser(getVersion(),factory,parser);
	}
	
	/**
	 * Gets the GML iteartor used to retrieve all the 
	 * features
	 * @param factory
	 * Factory to create the geometries
	 * @return
	 */
	public IGMLFeaturesIterator getFeaturesIterator() throws Exception{
		return featuresParser.getFeaturesReader();
	}		

	/**
	 * @return Returns the version.
	 */
	public String getVersion(){
		return schemasManager.getVersion();
	}

	/**
	 * @return Returns the extent.
	 */	
	public Rectangle2D getExtent(){
		return featuresParser.getExtent().getExtent();
	}
	
	/**
	 * @return Returns the SRS
	 */	
	public String getSRS(){
		return featuresParser.getExtent().getSrs();
	}
}
